// bingo2 for PLB (C) 2024 Shimeji Sola
// Common classes and routines

// Global constants, variables, and utility functions, used by both client and lottery
class GameEnvironment {
  constructor() {
    this.baseURL = 'https://localhost:8086/bingo2/';
    this.imageFolder = 'images/';
    this.gameFontFamily = '"Poetsen One"';
    this.defaultFont = { fontFamily: this.gameFontFamily, color: "#000", fontSize: "48px" };
    this.sceneAfterLoad = null;
    
    this.colors = ['r', 'b', 'g', 'y', 'p'];
    this.columnColor      = {'r': 0xff0000, 'b': 0x0000ff, 'g': 0x00cc00, 'y': 0xaaaa00, 'p': 0xcc00ff};
    this.columnColorLight = {'r': 0xffeeee, 'b': 0xeeeeff, 'g': 0xeeffee, 'y': 0xffffee, 'p': 0xf7eeff};
    this.columnColorFore  = {'r': 0xffdddd, 'b': 0xddddff, 'g': 0xddffdd, 'y': 0xffffdd, 'p': 0xeeddff};
    this.columnColorBack  = {'r': 0x880000, 'b': 0x000088, 'g': 0x006600, 'y': 0x666600, 'p': 0x660088};
    
    this.sizes = ['s', 'h', 'v', 'l'];
    this.cardStyles = {
      's': {'sx':  56, 'sy':  56, 'sw': 400, 'sh': 400, 'dw': 200, 'dh': 200},
      'h': {'sx':   0, 'sy': 142, 'sw': 512, 'sh': 228, 'dw': 450, 'dh': 200},
      'v': {'sx': 142, 'sy':   0, 'sw': 228, 'sh': 512, 'dw': 200, 'dh': 450},
      'l': {'sx':  31, 'sy':  31, 'sw': 450, 'sh': 450, 'dw': 450, 'dh': 450}};
    this.baseWidth = 20;

    this.numPerColumn = 10;
    this.numberImages = [];
    this.numLoadingImages = 0;
    this.roomExpire = 10800;
  }

  // returns a color for rainbow flash
  rainbow(value) {
    let ph = Math.floor(value * 6);
    let re = Math.floor((value * 6 - ph) * 255);
    let r = (ph == 4) ? re : (ph == 5 || ph == 0) ? 255: (ph == 1) ? 255 - re : 0;
    let g = (ph == 0) ? re : (ph == 1 || ph == 2) ? 255: (ph == 3) ? 255 - re : 0;
    let b = (ph == 2) ? re : (ph == 3 || ph == 4) ? 255: (ph == 5) ? 255 - re : 0;
    return r * 0x10000 + g * 0x100 + b;
  }
}

var env = new GameEnvironment();

/**********************************************************************/
// clickable button
class GameButton extends Phaser.GameObjects.Layer {
  constructor(scene, str, x, y, w, h, isCircle, callback = console.log) {
    super(scene);

    // base of the button
    const base = scene.add.graphics({x: x, y: y});
    base.lineStyle(4, 0xc080ff);
    base.fillStyle(0xe0c0ff);
    if (isCircle) {
      base.fillCircle(w / 2, w / 2, w / 2 - 2);
      base.strokeCircle(w / 2, w / 2, w / 2 - 2);
      const shape = new Phaser.Geom.Circle(w / 2, w / 2, w / 2);
      base.setInteractive(shape, Phaser.Geom.Circle.Contains);
    } else {
      base.fillRoundedRect(2, 2, w - 4, h - 4, w / 8);
      base.strokeRoundedRect(2, 2, w - 4, h - 4, w / 8);
      const shape = new Phaser.Geom.Rectangle(0, 0, w, h);
      base.setInteractive(shape, Phaser.Geom.Rectangle.Contains);
    }

    // text of the button
    const textFont = { fontFamily: env.gameFontFamily, color: "#306", fontSize: Math.min(80, h * 0.75) };
    const text = scene.add.text(x + w / 2, y + h / 2, str, textFont).setOrigin(0.5);
    text.setStroke('#c080ff', 2);

    // highlight
    this.eff = scene.add.graphics({x: x, y: y});
    this.effStyle = {w: w, h: h, isCircle: isCircle};
    this.updateEffect(0);

    // add created game objects to the layer
    this.add(base);
    this.add(text);
    this.add(this.eff);
    this.enabled = true;
    this.str = str;
    this.callback = callback;

    // register event listener
    base.on('pointerover', (e, g) => {
      if (this.enabled) {
        this.updateEffect(0.5);
      }
    });
    base.on('pointerout', (e, g) => {
      if (this.enabled) {
        this.updateEffect(0);
      }
    });
    base.on('pointerdown', (e, g) => {
      if (this.enabled) {
        this.callback(this.str);
      }
    });
  }

  // get the button lighter(dim > 0) or darker (dim < 0)
  updateEffect(dim) {
    this.eff.clear();
    this.eff.setAlpha(Math.abs(dim));
    this.eff.fillStyle((dim >= 0) ? 0xffffff : 0);
    if (this.effStyle.isCircle) {
      const r = this.effStyle.w / 2;
      this.eff.fillCircle(r, r, r);
    } else {
      const w = this.effStyle.w;
      const h = this.effStyle.h;
      this.eff.fillRoundedRect(0, 0, w, h, w / 8);
    }
  }

  // enable click events
  enable() {
    this.enabled = true;
    this.updateEffect(0);
  }

  // disable click events
  disable() {
    this.enabled = false;
    this.updateEffect(-0.5);
  }
}

/**********************************************************************/
// cell graphic layer
class BingoCell extends Phaser.GameObjects.Layer {
  constructor(scene, num, size, x, y, scale = 1, multiNum = true) {
    super(scene);

    // instance variables
    this.size = size;
    this.scale = scale;
    this.multiNum = multiNum;
    this.hit = false;
    this.nums = this.getNums(num);
    this.offsetX = 0;
    this.offsetY = 0;

    // local variables
    let style = env.cardStyles[size];
    let h = style['dh'] * scale;
    let w = style['dw'] * scale;
    let bw = env.baseWidth * scale;
    let col = env.colors[Math.floor((num - 1) / env.numPerColumn)];
    let font = {
      fontFamily: env.gameFontFamily,
      color: this.getColorStr(env.columnColorFore[col]),
      fontSize: 88 * scale };
    let numStr = this.nums.join('+');
    
    // colored base of the cell
    let baseOuter = scene.add.rectangle(x, y, w + bw * 2, h + bw * 2, 0).setOrigin(0);
    this.base     = scene.add.rectangle(x + 2 * scale, y + 2 * scale, w + bw * 2 - 4 * scale,
      h + bw * 2 - 4 * scale, 0xffffff).setOrigin(0);
    let baseInner = scene.add.rectangle(x + bw - 2, y + bw - 2, w + 4, h + 4, 0).setOrigin(0);
    
    // character icon
    this.bg = scene.add.image(x + bw, y + bw, 'back_' + size + col).setOrigin(0).setScale(scale);
    this.icon = scene.add.image(x + bw, y + bw, 'icon_' + num.toString()).setOrigin(0).setScale(scale);
    this.icon.setVisible(scene.textures.exists('icon_' + num.toString()));
    this.iconMask = scene.add.rectangle(x + bw, y + bw, w, h, 0xffffff).setOrigin(0).setAlpha(0);
    this.icon.setOrigin(style['sx'] / 512, style['sy'] / 512);
    this.icon.setScale(w / style['sw'], h / style['sh']);
    this.icon.setMask(this.iconMask.createGeometryMask());
    let frame  = scene.add.image(x + bw, y + bw, 'frame_' + size).setOrigin(0).setAlpha(0.3).setScale(scale);
    this.numBack = scene.add.text(x + bw + w - 18 * scale, y + bw + h, numStr, font).setOrigin(1);
    this.numBack.setStroke('#fff', 12 * scale);
    this.numFore = scene.add.text(x + bw + w - 20 * scale, y + bw + h - 2 * scale, numStr, font).setOrigin(1);
    this.numFore.setStroke(this.getColorStr(env.columnColorBack[col]), 8 * scale);
    
    // add created game objects to the layer
    this.add(baseOuter);
    this.add(this.base);
    this.add(baseInner);
    this.add(this.bg);
    this.add(this.icon);
    this.add(frame);
    this.add(this.numBack);
    this.add(this.numFore);
    this.add(this.iconMask);
  }

  // set the color of the base surrounding icon
  setBaseColor(col) {
    this.base.setFillStyle(col);
  }

  // set the alpha value of the icon
  setIconAlpha(num) {
    this.iconMask.setAlpha(1 - num);
  }

  // set an offset of position, origin is where the layer is instanciated
  setOffset(x, y) {
    let dx = x - this.offsetX;
    let dy = y - this.offsetY;
    let c = this.getChildren();
    for (let i = 0; i < c.length; i++) {
      c[i].x += dx;
      c[i].y += dy;
    }
    this.offsetX = x;
    this.offsetY = y;
  }

  // set the number of the cell
  setNum(num) {
    let col = env.colors[Math.floor((num - 1) / env.numPerColumn)];
    this.nums = this.getNums(num);
    let numStr = this.nums.join('+');
    this.bg.setTexture('back_' + this.size + col);
    this.icon.setTexture('icon_' + num.toString());
    this.icon.setVisible(this.scene.textures.exists('icon_' + num.toString()));
    this.numBack.setText(numStr);
    this.numFore.setText(numStr);
    this.numFore.setFill(this.getColorStr(env.columnColorFore[col]));
    this.numFore.setStroke(this.getColorStr(env.columnColorBack[col]), 8 * this.scale);
  }

  // get corresponding number(s) to the cell
  getNums(num) {
    let nn = Math.floor((num - 1) / 2) * 2 + 1;
    return (this.size == 'l' && this.multiNum) ? [nn, nn + 1] : [num];
  }

  // check if input number belongs to the cell
  checkNum(num) {
    return this.nums.includes(num);
  }

  // encode a color code to hexadecimal string
  getColorStr(col) {
    return '#' + col.toString(16).padStart(6, '0');
  }
}

/**********************************************************************/
// room information loader and parser
class RoomInfoLoader {
  constructor(room, action, callback) {
    let tm = Math.floor(Date.now() / 1000);
    this.req = new XMLHttpRequest();
    let loadFunc = this.fileLoaded.bind(this);
    let errorFunc = this.httpError.bind(this);
    this.req.addEventListener("load", loadFunc);
    this.req.addEventListener("error", errorFunc);
    if (action == "read") {
      this.req.open("GET", env.baseURL + "rooms/" + room + ".txt?" + tm.toString());
      this.req.setRequestHeader("Cache-Control", "no-cache");
      this.req.send();
    } else {
      this.req.open("POST", env.baseURL + "server.rb");
      this.req.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
      this.req.setRequestHeader("Cache-Control", "no-cache");
      this.req.send("room=" + room + "&action=" + action);
    }
    this.room = Number.parseInt(room, 10);
    this.valid = false;
    this.message = "";
    this.drawn = [];
    this.callback = callback;
  }

  fileLoaded(event) {
    let tm = Math.floor(Date.now() / 1000);
    if (Math.floor(this.req.status / 100) != 2) {
      this.message = (this.req.status == 404) ? "Room Not Open" : 
                     (this.req.status == 400) ? this.req.message : this.req.status.toString() + " Error";
    } else {
      let info = this.req.responseText.split(/\n|\r|\n\r/);
      info = info.map((x) => x.split(" ").map((y) => Number.parseInt(y)));
      this.valid = (info[0].length == 2 && info[0][0] == this.room && info[0][1] >= tm - env.roomExpire);
      if (! this.valid)
        this.message = "Room Not Open";
      for (let i = 1; i < info.length; i++) {
        if (info[i].length != 2)
          continue;
        if (info[i][0] == 0) {
          this.valid = false;
          this.message = "Room Closed";
        }
        this.drawn.push({num: info[i][0], tm: info[i][1] + info[0][1]});
      }
    }
    this.callback(this);
  }

  httpError(event) {
    this.message = "Connection Error";
    this.callback(this);
  }
}

/**********************************************************************/
// loading screen, wait until all of the icon images are loaded
class LoadScreen extends Phaser.Scene {
  constructor( ...args ) {
    super({ key: 'loading', ...args });
  }

  create() {
    let c = this.cameras.main.width / 2;
    this.add.text(c, 400, 'Loading images...', env.defaultFont).setOrigin(0.5, 0);
    this.add.rectangle(c - 310, 490, 620, 120).setOrigin(0).setStrokeStyle(1, 0);
    this.progressBar = this.add.rectangle(c - 300, 500, 1, 100, 0x330066).setOrigin(0);
    this.progressText = this.add.text(c, 650, '0 %', env.defaultFont).setOrigin(0.5, 0);

    // if images are not loaded, initiate loader threads
    if (env.numberImages.length == 0) {
      for (let i = 0; i < env.colors.length; i++) {
        for (let j = 0; j < env.numPerColumn; j++) {
          let index = i * env.numPerColumn + j + 1;
          let newImage = new Image();
          this.loadImagePromise(newImage, env.imageFolder + index.toString())
            .then((obj) => { env.numLoadingImages--; });
          env.numberImages.push(newImage);
        }
      }
      env.numLoadingImages = env.colors.length * env.numPerColumn;
    } else {
      // if images have been loaded, skip to the next screen
      this.scene.start(env.sceneAfterLoad);
    }
  }

  // a loader thread
  async loadImagePromise(img, src) {
    return new Promise((resolve, reject) => {
      img.onload = () => { resolve(img); };
      img.onerror = () => {
        img.onerror = () => { img.onerror = null; resolve(img); };
        img.src = img.src.substring(0, img.src.length - 4) + ".jpg";
      };
      img.src = src + ".png";
    });
  }

  // register a background image to collection of layers
  createIconBackground(size, col) {
    let style = env.cardStyles[size];
    let h = style['dh'];
    let w = style['dw'];

    let g = this.add.graphics();
    let mask = this.add.rectangle(0, 0, w, h, 0).setVisible(false);
    g.setMask(mask.createGeometryMask());

    g.fillStyle(env.columnColorFore[col]);
    g.fillRect(0, 0, w, h);
    g.lineStyle(8, env.columnColorLight[col]);
    for (let i = 20; i < h + w; i += 40) {
      g.beginPath();
      g.moveTo((i < h) ? 0 : i - h, (i < h) ? i : h);
      g.lineTo((i < w) ? i : w, (i < w) ? 0 : i - w);
      g.closePath();
      g.strokePath();
    }
    g.generateTexture('back_' + size + col, w, h);
    g.destroy();
  }

  // register a foreground image to collection of layers
  createIconFrame(size) {
    let style = env.cardStyles[size];
    let h = style['dh'];
    let w = style['dw'];

    let g = this.add.graphics();
    
    g.lineStyle(8, 0);
    g.beginPath();
    g.moveTo(w - 4, 8);
    g.lineTo(w - 4, h);
    g.moveTo(w - 4, h - 4);
    g.lineTo(8, h - 4);
    g.closePath();
    g.strokePath();
    g.lineStyle(8, 0xffffff);
    g.beginPath();
    g.moveTo(w, 4);
    g.lineTo(0, 4);
    g.moveTo(4, 4);
    g.lineTo(4, h);
    g.closePath();
    g.strokePath();

    g.generateTexture('frame_' + size, w, h);
    g.destroy();
  }

  // register an icon image for preview to collection of layers
  createIconPreview(num) {
    let col = env.colors[Math.floor((num - 1) / env.numPerColumn)];
    let style = env.cardStyles['s'];
    let img = document.createElement('canvas');
    img.width = style['dw'];
    img.height = style['dh'];
    let ctx = img.getContext('2d');

    let back = this.textures.get('back_s' + col).getSourceImage();
    ctx.drawImage(back, 0, 0);
    let icon = env.numberImages[num - 1];
    if (icon.width >= 4) {
      ctx.drawImage(icon, style['sx'], style['sy'], style['sw'], style['sh'], 0, 0, style['dw'], style['dh']);
    }
    let frame = this.textures.get('frame_s').getSourceImage();
    ctx.globalAlpha = 0.5;
    ctx.drawImage(frame, 0, 0);
    this.textures.addCanvas('preview_' + num.toString(), img);
  }

  // register a back image of icon for preview to collection of layers
  createBackPreview(col) {
    let style = env.cardStyles['s'];
    let img = document.createElement('canvas');
    img.width = style['dw'];
    img.height = style['dh'];
    let ctx = img.getContext('2d');

    let back = this.textures.get('back_s' + col).getSourceImage();
    ctx.drawImage(back, 0, 0);
    ctx.fillStyle = '#0004';
    ctx.fillRect(0, 0, style['dw'], style['dh']);
    this.textures.addCanvas('preview_b' + col, img);
  }

  update() {
    let p = Math.round((1 - env.numLoadingImages / env.numberImages.length) * 100);
    this.progressBar.setSize(p * 6, this.progressBar.height);
    this.progressText.setText(p.toString() + " %");

    // if all of the images are loaded, create and register required layers
    if (env.numLoadingImages == 0) {
      for (let i = 0; i < env.numberImages.length; i++)
        if (env.numberImages[i].width >= 4)
          this.textures.addImage("icon_" + (i + 1).toString(), env.numberImages[i]);
      for (let i = 0; i < env.sizes.length; i++) {
        for (let j = 0; j < env.colors.length; j++) {
          this.createIconBackground(env.sizes[i], env.colors[j]);
        }
        this.createIconFrame(env.sizes[i]);
      }
      for (let i = 0; i < env.numberImages.length; i++)
        this.createIconPreview(i + 1);
      for (let i = 0; i < env.colors.length; i++)
        this.createBackPreview(env.colors[i]);
      this.scene.start(env.sceneAfterLoad); // then proceed to the next screen
    }
  }
}

/**********************************************************************/