#!/usr/bin/env ruby
# bingo 2 for PLB (C) 2024 Shimeji Sola
# A bingo server, POSTed from the lottery client

require "cgi"
STATUS_STRING = {
  200 => "200 OK",
  400 => "400 Bad Request",
  405 => "405 Method Not Allowed"
}
ACTIONS = ['open', 'draw', 'close']
ROOM_DIR = __dir__ + '/rooms/'
NUM_MAX = 50
ROOM_EXPIRE = 10800
DRAW_DELAY = 12

# check if request is valid
def check_request(params)
  room = params['room'][0] # Room Code 
  v_room = "%04d" % (room.to_i(10))
  return nil if room != v_room
  action = params['action'][0] # action
  return nil if ! ACTIONS.include?(action)
  return [room, action]
end

# read room information
def get_room_info(room, tm)
  room_path = ROOM_DIR + room + '.txt'
  result = Hash.new
  result['valid'] = false
  return result if ! File.exist?(room_path)                       # room does not exist

  body = File.open(room_path){|f| f.read }
  info = body.split(/\R/).map{|x| x.split(' ').map{|y| y.to_i(10) } }
  return result if ! info[0]    || info[0][0] != room.to_i        # room code does not match
  return result if ! info[0][1] || info[0][1] + ROOM_EXPIRE < tm  # room has been expired
  result['body'] = body
  result['open'] = info[0][1]
  result['valid'] = true
  result['rest'] = (1..NUM_MAX).to_a
  info[1..-1].each do |x|
    next if x.length != 2
    if x[0] == 0
      result['valid'] = false                                     # room has been closed
    else 
      result['rest'].delete(x[0])
    end
  end
  return result
end

# write (append) room information
def put_room_info(room, num, tm, body = nil)
  line = num.to_s + ' ' + tm.to_s
  room_path = ROOM_DIR + room + '.txt'
  File.open(room_path, (body) ? 'a' : 'w') do |f|
    f.flock(File::LOCK_EX)
    f.puts(line)
  end
  return (body || "") + line + "\n"
end

# main routine that returns HTTP status and message
def server_main(cgi)
  return [405, 'Only POST method is allowed'] if cgi.request_method != "POST"

  room, action = check_request(cgi.params)
  return [400, 'Request is invalid']          if ! room

  tm = Time.now.to_i
  info = get_room_info(room, tm)

  if action == 'open'
    if info['valid']
      return [200, info['body']]                       # simply return the current room
    else
      return [200, put_room_info(room, room.to_i, tm)] # initialize the room by writing header
    end
  elsif action == 'draw'
    if info['valid'] 
      return [400, 'All numbers are drawn'] if info['rest'].empty?
      num = info['rest'].sample
      d_tm = tm - info['open'] + DRAW_DELAY
      return [200, put_room_info(room, num, d_tm, info['body'])]
    else
      return [400, 'The room has already been closed']
    end
  else # action == 'close'
    if info['valid'] 
      return [200, put_room_info(room, 0, 0, info['body'])]
    else
      return [400, 'The room has already been closed']
    end
  end
end

### calling main routine ###
cgi = CGI.new
status_code, message = server_main(cgi)
cgi.out({"status" => STATUS_STRING[status_code], "type" => "text/plain"}){ message }